from slai.loaders.torch_loader import TorchLoader
from slai.loaders.sklearn_loader import SklearnLoader
from slai.modules.model_saver import ValidModelFrameworks
from slai.exceptions import UnsupportedModelFormat


class ModelLoader:
    @classmethod
    def load_model(cls, *, model_artifact, deployment_instance_path, artifact_type):
        loaded_model = None

        if artifact_type == ValidModelFrameworks.Torch:
            loaded_model, inference_method_name = TorchLoader.load_model(
                model_artifact=model_artifact,
                deployment_instance_path=deployment_instance_path,
            )
        elif artifact_type == ValidModelFrameworks.Sklearn:
            loaded_model, inference_method_name = SklearnLoader.load_model(
                model_artifact=model_artifact,
                deployment_instance_path=deployment_instance_path,
            )
        else:
            raise UnsupportedModelFormat("invalid_artifact_type")

        return loaded_model, inference_method_name



import json
import subprocess
import os


class DataSet:
    def __init__(self, model, model_version):
        self.model_name = model["name"]
        self.model_id = model["id"]
        self.model_version = model_version
        self.s3_prefix = f"{self.model_name}/{self.model_version}/"

    def add_file(self, path, dir=None):
        pass

    def add_files(self, files, dir=None):
        pass

    def add_directory(self, path):
        pass

    def ls(self):
        pass

    def local_path(self):
        """
        Return the local path to the dataset
        """
        pass

    def changes(self):
        """
        Return local changes to dataset, to be used before calling
        """

    def save(self):
        """
        Back up dataset to s3
        """



import boto3

from slai.exceptions import InvalidAWSProfile, InvalidServicePrefix
from botocore.exceptions import ClientError, ProfileNotFound


class SSMClient:
    def __init__(self, *, aws_profile, aws_region):
        try:
            self.session = boto3.Session(
                profile_name=aws_profile, region_name=aws_region
            )
        except ProfileNotFound:
            raise InvalidAWSProfile("profile_not_found")

        self.sts_client = self.session.client("sts")
        self.ssm_client = self.session.client("ssm")

    def get_account_id(self):
        return self.sts_client.get_caller_identity()["Account"]

    def put_parameter(self, *, key, value):
        # TODO: add error handling
        response = self.ssm_client.put_parameter(
            Name=key,
            Value=value,
            Type="String",
            Overwrite=True,
        )
        return response

    def get_parameter(self, *, key):
        response = self.ssm_client.get_parameter(Name=key)
        value = response["Parameter"]["Value"]
        return value

    def list_parameters(self, *, prefix):
        parameters = []

        response = self.ssm_client.describe_parameters(MaxResults=10)
        next_token = response["NextToken"]

        while next_token:
            next_token = response.get("NextToken")
            if not next_token:
                break

            parameters.extend(response["Parameters"])
            response = self.ssm_client.describe_parameters(
                MaxResults=20, NextToken=next_token
            )

        parameter_names = []

        if prefix is None:
            raise InvalidServicePrefix("invalid_service_name")

        for param in parameters:
            if prefix in param["Name"]:
                parameter_names.append(param["Name"])

        return parameter_names



from slai.modules.parameters import from_config


def get_api_base_url():
    stage = from_config(
        key="STAGE",
        default="staging",
    )

    if stage == "local":
        base_url = from_config(
            key="BASE_URL",
            default="http://localhost:5000",
        )
    elif stage == "docker":
        base_url = "http://host.docker.internal:5000"
    else:
        base_url = from_config(
            key="BASE_URL",
            default=f"https://api.slai.io/{stage}",
        )

    return base_url



import re


from slai.clients.inference import get_inference_client
from slai.exceptions import InvalidModelURI
from slai.constants import MODEL_ROUTE_URI


class Model:
    def __init__(self, model_uri):
        self._parse_model_uri(model_uri)

        self.inference_client = get_inference_client(
            model_name=self.model_name,
            project_name=self.project_name,
            model_version_name=self.model_version_name,
        )

    def _parse_model_uri(self, model_uri):
        m = re.match(MODEL_ROUTE_URI, model_uri)
        if not m:
            raise InvalidModelURI("invalid_model_route")

        project_name = m.group(1)
        model_name = m.group(2)
        model_version_name = m.group(3)

        if model_version_name == "":
            model_version_name = None

        self.project_name = project_name
        self.model_name = model_name
        self.model_version_name = model_version_name

    def __call__(self, **inputs):
        return self.inference_client.call(payload=inputs)

    def info(self):
        pass



import io
import base64
import inspect
import json
import importlib.util
import sys

from importlib import import_module
from slai.loaders import base_loader


class TorchLoader(base_loader.BaseLoader):
    @classmethod
    def load_model(cls, model_artifact, deployment_instance_path):
        model_artifact = json.loads(model_artifact)
        model_artifact_model_class_source = base64.b64decode(
            model_artifact["class_source"]
        ).decode("utf-8")

        model_artifact_model_state_dict = base64.b64decode(model_artifact["state_dict"])

        # write model to deployment instance environment
        model_path = f"{deployment_instance_path}/model.py"
        with open(model_path, "w") as f_out:
            f_out.write(model_artifact_model_class_source)

        # dynamically load model module
        spec = importlib.util.spec_from_file_location("model", model_path)
        model_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(model_module)
        sys.modules["model"] = model_module

        model_object = None
        for name, obj in inspect.getmembers(model_module):
            if inspect.isclass(obj):
                model_object = obj
                break

        # instantiate model instance
        loaded_model = model_object()

        _torch = import_module("torch")
        model_state_dict_binary = io.BytesIO(model_artifact_model_state_dict)
        _state_dict = _torch.load(model_state_dict_binary)

        loaded_model.load_state_dict(_state_dict)
        loaded_model.eval()

        inference_method_name = None
        return loaded_model, inference_method_name
